/*********************************************************************

  Boat.java	(C) Kenneth J. Turner, Mark A. Jennings	01/12/01, V1.3

  This class holds the information for each boat from state/histories
  to how to draw itself. The various methods service the
  RulesResolver and the main "SailRace" objects.

  KJT 18/08/01:

    o Minor tidy-up of code

    o Change in positioning of text

    o Overlap/windward calculation rewritten

    o Proximity check added

    o qsin corrected for argument < 0 or > 359

    o qsin/qcos calculated once

  KJT 01/12/01:

    o Colour added

    o Scenario change monitored

    o Boat limits calculated

    o Boat proximity calculated more accurately

*********************************************************************/

import waba.ui.*;
import waba.fx.*;
import waba.sys.*;
import waba.util.*;
// import extra.util.Maths;

public class Boat {
  // miscellaneous constants
  private static final Color hullColour =	// hull colour
    new Color (255, 176, 224);
  public static final int proximityTolerance = 6;

  // status constants
  public static final int UNKNOWN = -1;
  public static final int ABOVE_PROPER_COURSE = 0;
  public static final int BELOW_PROPER_COURSE = 1;
  public static final int FINISHING = 2;
  public static final int GYBING = 3;
  public static final int STARTING = 4;
  public static final int TACKING_TO_PORT = 5;
  public static final int TACKING_TO_STARBOARD = 6;
  public static final int WAS_AHEAD = 7;
  public static final int WAS_OVERLAPPED = 8;
  public static final int CLEAR_STATUS = 9;

  public static final int NUM_INTENTS = 10;

  // tack constants
  public static final int PORT = 0;
  public static final int STARBOARD = 1;

  // state variables
  private Vector pastOverlap;
  private boolean[] stateFlags;
  private boolean proximity = false;
  private boolean invisible = false;
  private int xPos;
  private int yPos;
  private int heading;
  private int sailPos;
  private int boatTack;
  private int maxBoats;
  private String boatName;
  private boolean badTack = false;
  private boolean luffing = false;
  private boolean sailingByLee = false;

  static int xPoints[] = { 5, 6, 4, 2, 0, -2, -4, -6, -5, 5 };
  static int yPoints[] = { 15, 0, -4, -6, -8, -6, -4, 0, 15, 15};
  private int points = xPoints.length;
  private int xHullPoints[] = new int[points];
  private int yHullPoints[] = new int[points];
  private int corners[][] = new int[4][2];
  private int xTextPoint;
  private int yTextPoint;
  private int coeff;
  private int windCent = 0;

  // the algorithmic trig approximation methods courtesy of WabaSoft

  static float sin[] = {
    0.0f, 0.17364818f, 0.34202015f, 0.5f,
    0.64278764f, 0.76604444f, 0.8660254f, 0.9396926f,
    0.9848077f, 1.0f, 0.9848077f, 0.9396926f,
    0.8660254f, 0.76604444f, 0.64278764f, 0.5f,
    0.34202015f, 0.17364818f, 0.0f, -0.17364818f,
    -0.34202015f, -0.5f, -0.64278764f, -0.76604444f,
    -0.8660254f, -0.9396926f, -0.9848077f, -1.0f,
    -0.9848077f, -0.9396926f, -0.8660254f, -0.76604444f,
    -0.64278764f, -0.5f, -0.34202015f, -0.17364818f,
    0.0f
  };

  // constructor: in normal runtime use the overlaps parameter is not needed.
  // This is used when "resurrecting" a previously saved boat.

  public Boat (
    int x, int y, int heading, int windDir, int tack, String name,
     Vector overlaps) {

    SailRule.scenarioChange = true;		// note scenario change
    stateFlags = new boolean[NUM_INTENTS];
    for (int i = 0; i < NUM_INTENTS; i++)
      stateFlags[i] = false;
    boatName = name;
    xPos = x;
    yPos = y;
    this.heading = heading % 360;
    boatTack = tack;
    setWind (windDir, heading, boatTack);	// heading and tack
    int xdiff = 0;
    if (overlaps == null)
      pastOverlap = new Vector ();
    else
      pastOverlap = overlaps;
    setCorners ();
  }

  // return true if the boat is above close-hauled (heading less than 40
  // degrees relative to the wind)

  public boolean aboveCloseHauled () {
    return (windCent <= 40 || windCent >= 320);
  }

  // return true if the tack choice is valid

  public boolean boatOK () {
    return (!badTack);
  }

  // remove the selected state flag choice

  public void cancelIntention (int intent) {
    SailRule.scenarioChange = true;		// note scenario change
    stateFlags[intent] = false;
  }

  // convert a flag to text

  private String convertIntent (int i) {
    String result = "";
    switch (i) {
      case ABOVE_PROPER_COURSE:
        result = "AC"; break;
      case BELOW_PROPER_COURSE:
        result = "BC"; break;
      case FINISHING:
        result = "FN"; break;
      case GYBING:
        result = "GY"; break;
      case STARTING:
        result = "ST"; break;
      case TACKING_TO_PORT:
        result = "TP"; break;
      case TACKING_TO_STARBOARD:
        result = "TS"; break;
      case WAS_AHEAD:
        result = "WA"; break;
      case WAS_OVERLAPPED:
        result = "WO"; break;
    }
    return (result);
  }

  // change boat to other tack

  public void changeTack (int wind) {
    if (boatTack == PORT)
      boatTack = STARBOARD;
    else
      boatTack = PORT;
    setWind (wind, heading, boatTack);
  }

  // draw a boat

  public void drawBoat (Graphics g) {
    float sinHeading = qsin (heading);
    float cosHeading = qcos (heading);
    float sinSail = qsin (sailPos);
    float sinSailP7 = qsin (sailPos + 7);
    float sinSailM7 = qsin (sailPos - 7);
    float cosSail = qcos (sailPos);
    float cosSailP7 = qcos (sailPos + 7);
    float cosSailM7 = qcos (sailPos - 7);

    if (!invisible) {
      int horizPos = xTextPoint;
      int vertPos = yTextPoint + 2;

      g.setForeColor (Color.BLACK);
      if (badTack && !luffing)
	g.drawText ("Sail impossible!", xPos, yPos);
      else {
	// draw hull
        for (int i = 0; i < points; i++) {
	  xHullPoints[i] = xPos + (int) ((cosHeading * xPoints[i] + 0.5f) -
	      (sinHeading * yPoints[i] + 0.5f));
	  yHullPoints[i] = yPos + (int) ((sinHeading * xPoints[i] + 0.5f) +
	      (cosHeading * yPoints[i] + 0.5f));
        }
	g.setBackColor (hullColour);
	g.fillPolygon (xHullPoints, yHullPoints, points);
	g.setBackColor (SailRule.mapColour);
	g.drawPolygon (xHullPoints, yHullPoints, points);

        // draw sail
	if (!luffing)
	  g.drawLine (
	    xPos +
	      (int) ((cosSail * 0 + 0.5f) - (sinSail * 0 + 0.5f)),
	    yPos +
	      (int) ((sinSail * 0 + 0.5f) + (cosSail * 0 + 0.5f)),
	    xPos +
	      (int) ((cosSail * 0 + 0.5f) - (sinSail * -15 + 0.5f)),
	    yPos +
	      (int) ((sinSail * 0 + 0.5f) + (cosSail * -15 + 0.5f)));
	else {
	  g.drawLine (
	    xPos + (int) ((cosSail * 0 + 0.5f) - (sinSail * 0 + 0.5f)),
	    yPos + (int) ((sinSail * 0 + 0.5f) + (cosSail * 0 + 0.5f)),
	    xPos + (int) ((cosSail * 0 + 0.5f) - (sinSail * -15 + 0.5f)),
	    yPos + (int) ((sinSail * 0 + 0.5f) + (cosSail * -15 + 0.5f)));
	  g.drawLine (
	    xPos +
	      (int) ((cosSail * 0 + 0.5f) - (sinSail * 0 + 0.5f)),
	    yPos +
	      (int) ((sinSail * 0 + 0.5f) + (cosSail * 0 + 0.5f)),
	    xPos +
	      (int) ((cosSailP7 * 0 + 0.5f) - (sinSailP7 * -15 + 0.5f)),
	    yPos +
	      (int) ((sinSailP7 * 0 + 0.5f) + (cosSailP7 * -15 + 0.5f)));
	  g.drawLine (
	    xPos + (int) ((cosSail * 0 + 0.5f) - (sinSail * 0 + 0.5f)),
	    yPos + (int) ((sinSail * 0 + 0.5f) + (cosSail * 0 + 0.5f)),
	    xPos + (int) ((cosSailM7 * 0 + 0.5f) - (sinSailM7 * -15 + 0.5f)),
	    yPos + (int) ((sinSailM7 * 0 + 0.5f) + (cosSailM7 * -15 + 0.5f)));
	}

	// name boat
 	if (proximity)
	  g.drawText (boatName + "*", horizPos, vertPos);
	else
	  g.drawText (boatName, horizPos, vertPos);
	vertPos += 2;

	// show flags and past overlaps
	String[] sFlags = new String[NUM_INTENTS];
	sFlags = showFlags ();
	int count = 0;
	while ((count < sFlags.length) && (sFlags[count] != null)) {
	  if (sFlags[count].equals ("WO")) {
	    for (int i = 0; i < pastOverlap.getCount (); i++)
	      g.drawText(sFlags[count] + " " + pastOverlap.get (i),
	        horizPos, vertPos += 9);
	  }
	  else
	    g.drawText(sFlags[count], horizPos, vertPos += 9);
	  count++;
	}
      }
    }
  }

  // return individual intent flag

  public boolean getFlag (int f) {
    return (0 <= f && f < NUM_INTENTS ? stateFlags[f] : false);
  }

  // get boat heading

  public int getHeading () {
    return (heading);
  }

  // return the boat corners enlarged by 1 unit;
  // first index 0 = back left, 1 = back right, 2 = front right,
  // 3 = front left (i.e. anticlockwise); second index is 0 = X, 1 = Y

  private int[][] getLimits () {
    int limits[][] = new int[4][2];
    float sinHeading = qsin (heading);
    float cosHeading = qcos (heading);
    limits[0][0] =
      (xPos + (int) ((cosHeading * -7 + 0.5f) - (sinHeading * 16 + 0.5f)));
    limits[0][1] =
      (yPos + (int) ((sinHeading * -7 + 0.5f) + (cosHeading * 16 + 0.5f)));
    limits[1][0] =
      (xPos + (int) ((cosHeading *  7 + 0.5f) - (sinHeading * 16 + 0.5f)));
    limits[1][1] =
      (yPos + (int) ((sinHeading *  7 + 0.5f) + (cosHeading * 16 + 0.5f)));
    limits[2][0] =
      (xPos + (int) ((cosHeading *  7 + 0.5f) - (sinHeading * -9 + 0.5f)));
    limits[2][1] =
      (yPos + (int) ((sinHeading *  7 + 0.5f) + (cosHeading * -9 + 0.5f)));
    limits[3][0] =
      (xPos + (int) ((cosHeading * -7 + 0.5f) - (sinHeading * -9 + 0.5f)));
    limits[3][1] =
      (yPos + (int) ((sinHeading * -7 + 0.5f) + (cosHeading * -9 + 0.5f)));
    return (limits);				// return boat limits
  }

  // get boat name

  public String getName () {
    return (boatName);
  }

  // return the past overlaps list

  public Vector getOverlaps () {
    return (pastOverlap);
  }

  // returns the four math-rotated corners

  public int[][] getPoints () {
    return (corners);
  }

  // return true if specified flag is true

  public boolean getState (int i) {
    return (stateFlags[i]);
  }

  // return the state flags array

  public boolean[] getStates () {
    return (stateFlags);
  }

  // get boat tack

  public int getTack () {
    return (boatTack);
  }

  // get boat X coordinate

  public int getX () {
    return (xPos);
  }

  // get boat Y coordinate

  public int getY () {
    return (yPos);
  }

  // return true if the boat is head to wind (heading 0 relative to wind,
  // +/- 5 degrees)

  public boolean headToWind () {
    return (windCent <= 5 || windCent >= 355);
  }

  // check if other boat is clear ahead of current boat

  public boolean isAhead (Boat b) {
    return (b.isAstern (this));
  }

  // check if given point is astern (on or to the right) of the vector from
  // point 0 to point 1

  public boolean isAstern (
   int x, int y, int x0, int y0, int x1, int y1) {
    boolean astern;
    int xDiff = x1 - x0;
    int yDiff = y1 - y0;

    if (xDiff != 0)
      astern = ((y - y0) * xDiff) >= ((x - x0) * yDiff);
    else if (yDiff > 0)
      astern = x <= x0;
    else
      astern = x > x0;
    return (astern);
  }

  // check if other boat is clear astern of current boat

  public boolean isAstern (Boat otherBoat) {
    int otherCorners[][] = otherBoat.getPoints ();
    int i;
    int x0 = corners[0][0];
    int y0 = corners[0][1];
    int x1 = corners[1][0];
    int y1 = corners[1][1];

    for (i = 0; i < 4 ; i++) {
      if (!isAstern (otherCorners[i][0], otherCorners[i][1],
             x0, y0, x1, y1))
        return (false);
    }
    return (true);
  }

  // check if current boat is in the band between the given two

  public boolean isBetween (Boat a, Boat b) {
    return (
      (this.isPort (a) && b.isPort (this)) ||
      (this.isPort (b) && a.isPort (this)));
  }

  // check if given boat is close to current boat

  public boolean isClose (Boat otherBoat) {
    return (this.isWithin (otherBoat) || otherBoat.isWithin (this));
  }

  // return true if the boat is close-hauled (heading 45 relative to wind,
  // +/- 5 degrees)

  public boolean isCloseHauled () {
    return (
      (40 <= windCent && windCent <= 50) ||
      (310 <= windCent && windCent <= 320));
  }

  // check if the given boat is luffing

  public boolean isLuffing () {
    return (luffing);
  }

  // check if given boat overlaps the current boat

  public boolean isOverlap (Boat otherBoat) {
    return (!this.isAstern (otherBoat) && !otherBoat.isAstern (this));
  }

  // check if other boat is to port of current one, i.e. if
  // the position of given boat is "astern" of the port side
  // line of the current boat

  public boolean isPort (Boat otherBoat) {
    int i;
    int headingDiff = heading - otherBoat.getHeading ();
    int otherCorners[][] = otherBoat.getPoints ();
    int x0, y0, x1, y1;			// port side line (3 to 0)
    int count1 = 0;			// other's points to port of current
    int count2 = 0;			// current's points to port of other

    if (headingDiff < 0)
      headingDiff += 360;
    x0 = corners[3][0]; y0 = corners[3][1];
    x1 = corners[0][0]; y1 = corners[0][1];
    for (i = 0; i < 4 ; i++) {
      if (isAstern (otherCorners[i][0], otherCorners[i][1],
             x0, y0, x1, y1))
        count1++;
    }
    x0 = otherCorners[3][0]; y0 = otherCorners[3][1];
    x1 = otherCorners[0][0]; y1 = otherCorners[0][1];
    for (i = 0; i < 4 ; i++) {
      if (isAstern (corners[i][0], corners[i][1],
             x0, y0, x1, y1))
        count2++;
    }
    return (count1 > count2 || (count1 == count2 && headingDiff < 180));
  }

  // check if other boat is windward of current one, i.e. if
  // the position of given boat is "astern" of the side line
  // carrying the sail of the current boat

  public boolean isWindward (Boat otherBoat) {
    int otherCorners[][] = otherBoat.getPoints ();
    int i;
    int x0, y0, x1, y1;

    if (boatTack == PORT) {
      x0 = corners[2][0]; y0 = corners[2][1];
      x1 = corners[1][0]; y1 = corners[1][1];
    }
    else {
      x0 = corners[0][0]; y0 = corners[0][1];
      x1 = corners[3][0]; y1 = corners[3][1];
    }
    for (i = 0; i < 4 ; i++) {
      if (isAstern (otherCorners[i][0], otherCorners[i][1],
             x0, y0, x1, y1))
        return (true);
    }
    return (false);
  }

  // check if given point is inside anticlockwise rectangle defined by corners

  public boolean isWithin (int x, int y, int corners[][]) {
    int x0 = corners[0][0];
    int y0 = corners[0][1];
    int x1 = corners[1][0];
    int y1 = corners[1][1];
    int x2 = corners[2][0];
    int y2 = corners[2][1];
    int x3 = corners[3][0];
    int y3 = corners[3][1];

    return (
      isAstern (x, y, x1, y1, x0, y0) &&
      isAstern (x, y, x2, y2, x1, y1) &&
      isAstern (x, y, x3, y3, x2, y2) &&
      isAstern (x, y, x0, y0, x3, y3));
  }

  // check if corners of other boat are within current boat corners
  // (both enlarged by distance tolerance)

  public boolean isWithin (Boat otherBoat) {
    int i;					// corner index
    int boatLimits1[][] = getLimits ();		// get current boat limits
    int boatLimits2[][] = otherBoat.getLimits (); // get other boat limits

    for (i = 0; i <= 3; i++) {			// go through limits
      if (isWithin (				// limits inside?
	    boatLimits2[i][0], boatLimits2[i][1], boatLimits1))
	return (true);				// report inside
    }
    return (false);				// report not inside
  }

  // check if mark is too close to current boat

  public boolean isWithin (Mark mark) {
    int i;					// corner index
    int boatLimits[][] = getLimits ();		// get current boat limits
    int markLimits[][] = mark.getLimits ();	// get mark limits

    for (i = 0; i <= 3; i++) {			// go through limits
      if (isWithin (				// limits inside?
	    markLimits[i][0], markLimits[i][1], boatLimits))
	return (true);				// report inside
    }
    return (false);				// report not inside
  }

  // check whether the boat is within two lengths (2 * 23) of the other boat
  // (i.e. the squared distance is less than 46 * 46)

  public boolean nearBoat (Boat otherBoat) {
    float distanceSquared;
    int boat1X;
    int boat1Y;
    int boat2X = otherBoat.getX ();
    int boat2Y = otherBoat.getY ();
    for (int i = 0; i < corners.length; i++) {
      boat1X = corners[i][0];
      boat1Y = corners[i][1];
      distanceSquared =
        (boat1X - boat2X) * (boat1X - boat2X) +
        (boat1Y - boat2Y) * (boat1Y - boat2Y);
      if (distanceSquared < 2116)
        return (true);
    }
    return (false);
  }

  // check whether the boat is within two lengths (2 * 23) of the given mark
  // (i.e. the squared distaince is less than 46 * 46)

  public boolean nearMark (Mark m) {
    float distanceSquared;
    int boatX;
    int boatY;
    int markX = m.getX ();
    int markY = m.getY ();
    for (int i = 0; i < corners.length; i++) {
      boatX = corners[i][0];
      boatY = corners[i][1];
      distanceSquared =
        (boatX - markX) * (boatX - markX) +
        (boatY - markY) * (boatY - markY);
      if (distanceSquared < 2116)
        return (true);
    }
    return (false);
  }

  // add a boat name to the past overlaps list

  public void overlapAdd (Boat b) {
    int i;
    String overlap;

    getFlag (WAS_OVERLAPPED);
    for (i = 0; i < pastOverlap.getCount (); i++) {
      overlap = (String) pastOverlap.get (i);
      if (overlap.equals (b.getName ()))
	return;
    }
    pastOverlap.add (b.getName ());
  }

   // clear out the list of past overlaps on request

  public Vector overlapDelete () {
    int i;
    int count;
    Vector overlaps = new Vector ();

    count = pastOverlap.getCount ();
    for (i = 0; i < count; i++)
      overlaps.add (pastOverlap.get (i));
    pastOverlap = new Vector ();
    return (overlaps);
  }

  // remove a specific boat from the history list

  public void overlapRemove (String name) {
    int i;
    String overlap;

    for (i = 0; i < pastOverlap.getCount (); i++) {
      overlap = (String) pastOverlap.get (i);
      if (overlap.equals (name))
	pastOverlap.del (i);
    }
    if (pastOverlap.getCount () == 0)
      stateFlags[WAS_OVERLAPPED] = false;
  }

  // change the name of any overlapped boat

  public void overlapRename (String oldName, String newName) {
    int i;
    String overlap;

    for (i = 0; i < pastOverlap.getCount (); i++) {
      overlap = (String) pastOverlap.get (i);
      if (overlap.equals (oldName))
	pastOverlap.set (i, newName);
    }
  }

 // calculate the cos of a value in degrees; the degrees walk
  // clockwise instead of counterclockwise to match the clock

  static float qcos (int deg)	{
    return qsin ((deg + 90) % 360);
  }

  // calculate the sin of a value in degrees (0...359)

  static float qsin (int deg) {
    // get the nearest sin we have in the table
    int d10 = deg / 10;
    if (d10 < 0)
      d10 += 36;
    else if (d10 > 35)
      d10 -= 36;
    float low = sin[d10];
    int weight = deg % 10;
    if (weight == 0)
      return (low);
    float high = sin[d10 + 1];
    float wf = (float) weight;
    return ((low * (10f - wf) + high * wf) / 10.0f);
  }

  // set boat coordinates

  public void setCoordinates (int x, int y) {
    SailRule.scenarioChange = true;		// note scenario change
    xPos = x;
    yPos = y;
    setCorners ();
  }

  // calculate the 'corners' used for overlap detection;
  // first index 0 = back left, 1 = back right, 2 = front right,
  // 3 = front left (i.e. anticlockwise); second index is 0 = X, 1 = Y

  private void setCorners () {
    float sinHeading = qsin (heading);
    float cosHeading = qcos (heading);
    corners[0][0] =
      (xPos + (int) ((cosHeading * -6 + 0.5f) - (sinHeading * 15 + 0.5f)));
    corners[0][1] =
      (yPos + (int) ((sinHeading * -6 + 0.5f) + (cosHeading * 15 + 0.5f)));
    corners[1][0] =
      (xPos + (int) ((cosHeading *  6 + 0.5f) - (sinHeading * 15 + 0.5f)));
    corners[1][1] =
      (yPos + (int) ((sinHeading *  6 + 0.5f) + (cosHeading * 15 + 0.5f)));
    corners[2][0] =
      (xPos + (int) ((cosHeading *  6 + 0.5f) - (sinHeading * -8 + 0.5f)));
    corners[2][1] =
      (yPos + (int) ((sinHeading *  6 + 0.5f) + (cosHeading * -8 + 0.5f)));
    corners[3][0] =
      (xPos + (int) ((cosHeading * -6 + 0.5f) - (sinHeading * -8 + 0.5f)));
    corners[3][1] =
      (yPos + (int) ((sinHeading * -6 + 0.5f) + (cosHeading * -8 + 0.5f)));

    if (0 < heading && heading <= 90) {
      xTextPoint = corners[1][0];
      yTextPoint = corners[1][1];
    }
    else if (90 < heading && heading <= 180) {
      xTextPoint = corners[2][0];
      yTextPoint = corners[2][1];
    }
    else if (180 < heading && heading <= 270) {
      xTextPoint = corners[3][0];
      yTextPoint = corners[3][1];
    }
    else {
      xTextPoint = corners[0][0];
      yTextPoint = corners[0][1];
    }
  }

  // set the boat heading

  public void setHeading (int heading, int wind) {
    SailRule.scenarioChange = true;		// note scenario change
    this.heading = heading;
    setWind (wind, heading, boatTack);
    setCorners ();
  }

  // set the states, checking that the choice is basically valid
  // given the boat's current state.

  public void setIntention (int intent) {
    SailRule.scenarioChange = true;		// note scenario change
    if (intent == ABOVE_PROPER_COURSE && stateFlags[BELOW_PROPER_COURSE])
      stateFlags[BELOW_PROPER_COURSE] = false;
    else if (intent == BELOW_PROPER_COURSE && stateFlags[ABOVE_PROPER_COURSE])
      stateFlags[ABOVE_PROPER_COURSE] = false;
    else if (intent == TACKING_TO_STARBOARD && stateFlags[TACKING_TO_PORT])
      stateFlags[TACKING_TO_PORT] = false;
    else if (intent == TACKING_TO_PORT && stateFlags[TACKING_TO_STARBOARD])
      stateFlags[TACKING_TO_STARBOARD] = false;
    else if (intent == STARTING && stateFlags[FINISHING])
      stateFlags[FINISHING] = false;
    else if (intent == FINISHING && stateFlags[STARTING])
      stateFlags[STARTING] = false;
    else if (intent == CLEAR_STATUS) {
      for (int i = 0; i < stateFlags.length; i++)
        stateFlags[i] = false;
     return;
    }
    stateFlags[intent] = true;
  }

  // cancel drawing of the boat for the view screen prior to a bat "re-spot"

  public void setInvisible () {
    invisible = true;
  }

   // set boat name

  public void setName (String name) {
    SailRule.scenarioChange = true;		// note scenario change
    boatName = name;
  }

  // set proximity if another boat is too close

  public void setProximity (boolean close) {
    proximity = close;
  }

  // set the entire state flag array according to the array passed in

  public void setStates (boolean[] b) {
    for (int i = 0; i < b.length; i++)
      stateFlags[i] = b[i];
  }

  // set boat tack

  public void setTack (int tack, int wind) {
    boatTack = tack;
    setWind (wind, heading, boatTack);
  }

  // set drawing of the boat for the view screen prior to a bat "re-spot"

  public void setVisible () {
    invisible = false;
  }

  // for setting and changing the wind direction and boat heading relative to
  // this boat's current settings.  This function works out the position
  // of the sail relative to the boat's heading and the wind direction.

  public void setWind (int wind, int heading, int boatTack) {
    // first the wind direction fix (back to the REAL way the wind is gauged:
    // where the wind is coming FROM):
    int windDir = (wind + 180) % 360;

    // work out the wind centre line angle, namely what the wind centre
    // line angle is once the boat's heading is accounted for
    windCent = 0;
    if (heading <= 180)
      windCent = ((180 - heading) + windDir) % 360;
    else
      windCent = ((180 - heading) + windDir + 360) % 360;

    // is the boat's sail "flapping"?
    if ((windCent < 45 && boatTack == STARBOARD) ||
	((windCent > 315 || windCent == 0) && boatTack == PORT))
      luffing = true;
    else
      luffing = false;

    // next work out if this is permissible with current tack STARBOARD:
    if (boatTack == STARBOARD) {
      // explicitly defined tack cannot support this sail position given wind
      if (windCent > 190)
	badTack = true;
      else {
	badTack = false;
	// apply the formula for the sail angle...
	sailPos = (int) ((windCent) * 0.55f - 8.32f);
	// check to see if the wind is blowing the sail
	// right out from behind:
	if (sailPos > 90)
	  sailPos = 90;
	// the final angle is worked out in accordance with
	// the sail being over the port side of the boat.
	sailPos = (heading + (180 + sailPos)) % 360;
	if ( windCent > 180)
	  sailingByLee = true;
	else
	  sailingByLee = false;
      }
    }
    // next work out if this is permissible with current tack PORT
    else if (boatTack == PORT) {
      if (windCent < 170 && (windCent !=0))
	badTack = true;
      else {
	badTack = false;
	// apply the formula:
	sailPos = (int) ((360 - windCent)*0.54f - 8.32f) % 360;
	if (windCent == 0)
	  sailPos = 0;
	else {
	  // fully extended sail:
	  if (sailPos > 90)
	    sailPos = 90;
	}
	// this time with the sail sits over the
	// starboard side of the boat:
	sailPos = heading - (180 + sailPos);
	if ( sailPos < 0)
	  sailPos += 360;
	if (windCent < 180)
	  sailingByLee = true;
	else
	  sailingByLee = false;
      }
    }
  }

  // hand back an array tailored to the length of only the true flags

  public String[] showFlags( ) {
    String[] Flags = new String[NUM_INTENTS];
    int fCount = 0;
    for (int i = 0;i<stateFlags.length;i++)
      if (stateFlags[i])
	Flags[fCount++] = convertIntent (i);
    return (Flags);
  }

  // detect if the boat passed in is stored in the internal past
  // overlaps Vector.

  public boolean wasOverlapped (Boat b) {
    String nameA = b.getName ();
    String nameB;
    for (int i = 0; i < pastOverlap.getCount (); i++) {
      nameB = (String) pastOverlap.get (i);
      if (nameA.equals (nameB)) {
	return (true);
      }
    }
    return (false);
  }
}
