/*********************************************************************

  SailRule.java	(C) Kenneth J. Turner, Mark A. Jennings	21/08/02 V1.5

  The main sailing program. Only one mark is supported by the rules resolver
  at this time.

  KJT 04/06/01:

    o ScenarioChange "b." to "Boat."

  KJT 16/08/01:

    o Tidy-up of code

    o All screens and user interface redesigned

    o Help screens added

    o Graphical editing added

    o Rule handling scenarioChange to array of indexes

  KJT 18/11/01:

    o Name scenarioChange to "SailRule"

    o "x" to exit map screen

    o Heading tolerance for mark drags to avoid rapid port/starboard change

    o Files in alphabetical order

  KJT 01/12/01:

    o Changes for Waba 2.0

    o Clearing via popup window

    o Warning of need to save on load or quit

  KJT 21/08/02:

    o Changes in widget backgrounds for SuperWaba 3.0 compatibility

  KJT 31/07/03:

    o Recompiled to include WinCE files

*********************************************************************/

import waba.fx.*;
import waba.io.*;
import waba.sys.*;
import waba.ui.*;
import waba.util.*;
import extra.io.*;
import extra.util.*;
import java.lang.*;
import jgui.ui.*;

public class SailRule extends MainWindow {
  // boat status code constants
  private static final String clr = "--: Clear (no status)";
  private static final String apc = "AC: Above proper course";
  private static final String bpc = "BC: Below proper course";
  private static final String fin = "FN: Finishing";
  private static final String gyb = "GY: Gybing";
  private static final String sta = "ST: Starting";
  private static final String tkp = "TP: Tacking to port";
  private static final String tks = "TS: Tacking to starboard";
  private static final String wad = "WA: Was ahead";
  private static final String wol = "WO: Was overlapped";

  // graphical boat edit mode constants
  public static final int NO_MODE = 0;
  public static final int HEADING_MODE = 1;
  public static final int PLACE_MODE = 2;

  // menus
  private static String mainMenu[] = {
    "SailRace", "About...", "Quit..."
  };
  private static String mapMenu[] = {
    "Map",
    "Add Boat           B", "Add Mark         M", "New Heading H",
    "New Place        P", "-",
    "Edit Boats        E", "Show Files        F", "Show Rules     R",
    "Map Exit           X", "-", "Clear...               C"
  };
  private final MenuBar mainBar =		// set up main menu
    new MenuBar (new String[][]{mainMenu});
  private final MenuBar mainMapBar =		// set up main and map menus
    new MenuBar (new String[][]{mainMenu, mapMenu});

  // message boxes
  private MessageBox aboutMessage = new MessageBox (
    "About SailRace",
    "SAILING RACE RULES FOR PDAs||" +
    "Kenneth J. Turner (maintainer)| <kjt@cs.stir.ac.uk>|" +
    "Mark A. Jennings (joint developer)| <markeyj@talk21.com>|" +
    "www.cs.stir.ac.uk/~kjt/software/|  pda/sailrace.html||" +
    "Version 1.5 (c) 31st July 2003|" +
    "GPL Licence: no fee, no warranty!||" +
    "Created with SuperWaba||http://www.superwaba.org/||" +
    "See the software license at|this URL, or email|" +
    "<guich@superwaba.org>",
    new String[] {"OK"});
  private MessageBox clearMessage = new MessageBox (
    "Clear Scenario",
    "Clear current|scenario without|saving it?",
    new String[] {"No", "Clear"});
  private MessageBox loadMessage = new MessageBox (
    "Load Scenario",
    "Save current scenario|before loading?",
    new String[] {"No", "Save"});
  private MessageBox quitMessage = new MessageBox (
    "Quit Program",
    "Save current scenario|before quitting?",
    new String[] {"No", "Save"});

  // miscellaneous constants
  public static final Color backgroundColour =	// background colour
    new Color (220, 255, 255);
  private static final int boatMax = 4;		// maximum boat count
  private static final String catalogueName =	// database name
    "SailRuleDB.SlRl.DATA";
  private static final Color borderColour =	// border colour
    new Color (80, 80, 128);
  public static final Color foregroundColour =	// foreground colour
    Color.BLACK;
  private static final int fileMax = 64;	// maximum file count
  private static final int headingTolerance = 2;
  public static final Color inputColour =	// input background colour
    backgroundColour.brighter ();
  private static final int markMax = 1;		// maximum mark count
  public static final Color mapColour =		// map background colour
    Color.WHITE;
  private static final int mapEdge = 18;	// closeness to map edge
  private static final int mapTolerance = 10;	// closeness to boat/mark
  private static final int pageMax = 5;		// maximum page count
  private static final int scrollIncrement =	// scroll bar increment
    132;
  public static final Color textColour =	// input text colour
    Color.BLACK;
  private static final Font textFont =		// default font
    new Font ("Helvetica", Font.PLAIN, 12);

  // page code constants
  public static final int NO_PAGE = 0;
  public static final int HELP_PAGE = 1;
  public static final int MAIN_PAGE = 2;
  public static final int START_PAGE = 3;
  public static final int MAP_PAGE = 4;
  public static final int BOAT_PAGE = 5;
  public static final int OVERLAPPED_PAGE = 6;
  public static final int MARK_PAGE = 7;
  public static final int MARK_DIR_PAGE = 8;
  public static final int EDIT_PAGE = 9;
  public static final int RULES_PAGE = 10;
  public static final int FILE_PAGE = 11;

  // state constants
  public static final int NORMAL = 0;
  public static final int DRAGGING = 1;
  public static final int LOADING = 2;
  public static final int QUITTING = 3;

  // common page widgets
  jgui.ui.TextArea commonText = new jgui.ui.TextArea ();
  Scroller commonScroller = new Scroller (commonText);
  private Label commonReport = new Label ("");
  private Button commonHelp = new Button ("Help");
  private Button commonDone = new Button ("Done");
  private MenuBar menuBar;

  // edit page widgets
  private Label editEditLabel = new Label ("Boat:");
  private Label editHeadingLabel = new Label ("Heading:");
  private Button editPrevious = new Button ("Prev");
  private Button editNext = new Button ("Next");
  private Edit editHeadingInput = new Edit ();
  private Label editNameLabel = new Label ("Name:");
  private Edit editNameInput = new Edit ();
  private ListBox editList = new ListBox ();
  private Button editRename = new Button ("Name");
  private Button editRedirect = new Button ("Head");
  private Button editReposition = new Button ("Place");
  private Button editDeleteBoat = new Button ("Delete");
  private ListBox editStatuses = new ListBox ();
  private Label editStatusLabel = new Label ("Click list above:");
  private Button editAddStatus = new Button ("Add");
  private Button editClearStatus = new Button ("Clear");

  // files page widgets
  private ListBox filesList = new ListBox ();
  private Label filesNameLabel = new Label ("Name:");
  private Edit filesNameInput = new Edit ();
  private Button filesLoad = new Button ("Load");
  private Button filesSave = new Button ("Save");
  private Button filesRename = new Button ("Name");
  private Button filesDelete = new Button ("Delete");

  // main page widgets
  private Label mainBoatLabel = new Label ("Boats:");
  private Label mainHeadingLabel = new Label ("Heading:");
  private Edit mainHeadingInput = new Edit ();
  private Label mainNameLabel = new Label ("Name: ");
  private Edit mainNameInput = new Edit ();
  private Radio mainPort = new Radio ("Port");
  private Radio mainStarboard = new Radio ("Stbd");
  private Button mainAddBoat = new Button ("Add");
  private Button mainEditBoats = new Button ("Edit");
  private Label mainMarkLabel = new Label ("Marks:");
  private Radio mainPortMark = new Radio ("Port");
  private Radio mainStarboardMark = new Radio ("Stbd");
  private Button mainAddMark = new Button ("Add");
  private Button mainDeleteMark = new Button ("Delete");
  private Label mainFileLabel = new Label ("Scenario:");
  private Button mainClear = new Button ("Clear");
  private Button mainFile = new Button ("File");
  private Button mainMap = new Button ("Map");
  private Button mainRules = new Button ("Rules");
  private Label mainWindLabel = new Label ("Wind:");

  // start page widgets
  private Label startOriginLabel = new Label ("Wind direction:");
  private Edit startOriginInput = new Edit ();

  // scenario variables
  private Descriptor blankDescriptor = new Descriptor ();
  private int boatCount;
  private Boat boatList[] = new Boat[boatMax];
  private String boatName;
  private Boat boatSelected;
  private int boatTack;
  private int boatHeading;
  private Case caseList;
  private Case casePointer;
  public static boolean scenarioChange;		// scenario changed
  private int editPoint;
  private int fileCount = 0;			// file count
  private String fileList[] = new String[fileMax]; // file names
  private Help help;
  private Graphics mapArea = new Graphics (this);
  private Image mapImage = new Image (160, 160);
  private Graphics mapBuffer = new Graphics (mapImage);
  private int markCount;
  private Mark markList[] = new Mark[markMax];
  private Mark markSelected;
  private int markTack;
  private int mode;
  private int state;				// program state
  private RulesResolver resolver;
  private int page, helpPage;			// current/help page
  private int pageCount;
  private int pageList[] = new int[pageMax];
  private boolean redraw;
  private boolean repositioning;
  private boolean usedRules[] = new boolean[RuleBook.NUM_RULES];
  private Wind wind;
  private int windDir;

  // main class

  public SailRule () {
    // general setup
    Vm.interceptSystemKeys (			// ignore system keys
      Vm.SK_ALL | Vm.SK_LAUNCH);
    this.setBackColor (backgroundColour);	// colour background for start
    page = START_PAGE;				// set start page
    initialiseScenario ();			// initialise scenario
    resolver = new RulesResolver ();		// create rules resolver object
    resolver.registerControl (this);		// register main class with it
    help = new Help ();				// create help object
    wind = new Wind ();				// create wind object

    // common page widgets
    commonText.setRect (0, 0, 148, 141);
    commonText.setBackForeColors (backgroundColour, textColour);
    commonText.setFont (textFont);
    commonScroller.setBackForeColors (backgroundColour, textColour);
    commonScroller.setIncrement (scrollIncrement);
    commonReport.setRect (5, 148, 80, 12);
    commonHelp.setRect (89, 148, 31, 12);
    commonDone.setRect (124, 148, 31, 12);

    // edit page widgets
    editEditLabel.setRect (0, 0, 30, 12);
    editPrevious.setRect (89, 1, 31, 12);
    editNext.setRect (124, 1, 31, 12);
    editHeadingLabel.setRect (5, 16, 40, 12);
    editHeadingInput.setRect (50, 16, 25, 14);
    editHeadingInput.setBackForeColors (inputColour, textColour);
    editHeadingInput.setValidChars ("0123456789");
    editNameLabel.setRect (79, 16, 31, 12);
    editNameInput.setRect (115, 16, 45, 14);
    editNameInput.setBackForeColors (inputColour, textColour);
    editList.setRect (5, 32, 150, 41);
    editList.setBackForeColors (inputColour, textColour);
    editRedirect.setRect (5, 75, 31, 12);
    editRename.setRect (39, 75, 31, 12);
    editReposition.setRect (89, 75, 31, 12);
    editDeleteBoat.setRect (124, 75, 31, 12);
    editStatuses.setRect (5, 90, 150, 41);
    editStatuses.setBackForeColors (inputColour, textColour);
    editStatuses.add (clr);
    editStatuses.add (apc);
    editStatuses.add (bpc);
    editStatuses.add (fin);
    editStatuses.add (gyb);
    editStatuses.add (sta);
    editStatuses.add (tkp);
    editStatuses.add (tks);
    editStatuses.add (wad);
    editStatuses.add (wol);
    editStatusLabel.setRect (5, 134, 70, 12);
    editAddStatus.setRect (89, 134, 31, 12);
    editClearStatus.setRect (124, 134, 31, 12);

    // files page widgets
    filesList.setRect (5, 5, 150, 104);
    filesList.setBackForeColors (inputColour, textColour);
    filesNameLabel.setRect (5, 112, 31, 12);
    filesNameInput.setRect (40, 112, 115, 14);
    filesNameInput.setBackForeColors (inputColour, textColour);
    filesLoad.setRect (5, 130, 31, 12);
    filesSave.setRect (40, 130, 31, 12);
    filesRename.setRect (89, 130, 31, 12);
    filesDelete.setRect (124, 130, 31, 12);

    // main page widgets
    mainBoatLabel.setRect (0, 0, 60, 12);
    mainHeadingLabel.setRect (5, 12, 40, 12);
    mainHeadingInput.setRect (50, 10, 25, 14);
    mainHeadingInput.setBackForeColors (inputColour, textColour);
    mainHeadingInput.setValidChars ("0123456789");
    mainNameLabel.setRect (79, 12, 31, 12);
    mainNameInput.setRect (115, 10, 45, 14);
    mainNameInput.setBackForeColors (inputColour, textColour);
    mainPort.setRect (5, 26, 32, 12);
    mainPort.setBackForeColors (backgroundColour, textColour);
    mainStarboard.setRect (40, 26, 32, 12);
    mainStarboard.setBackForeColors (backgroundColour, textColour);
    mainAddBoat.setRect (89, 26, 31, 12);
    mainEditBoats.setRect (124, 26, 31, 12);
    mainMarkLabel.setRect (0, 44, 60, 12);
    mainPortMark.setRect (5, 58, 32, 12);
    mainPortMark.setBackForeColors (backgroundColour, textColour);
    mainStarboardMark.setRect (40, 58, 32, 12);
    mainStarboardMark.setBackForeColors (backgroundColour, textColour);
    mainAddMark.setRect (89, 58, 31, 12);
    mainDeleteMark.setRect (124, 58, 31, 12);
    mainFileLabel.setRect (0, 76, 60, 12);
    mainClear.setRect (5, 88, 31, 12);
    mainFile.setRect (40, 88, 31, 12);
    mainMap.setRect (89, 88, 31, 12);
    mainRules.setRect (124, 88, 31, 12);
    mainWindLabel.setRect (0, 106, 50, 12);

    // start page widgets
    startOriginLabel.setRect (5, 132, 65, 12);
    startOriginInput.setRect (75, 132, 25, 14);
    startOriginInput.setBackForeColors (inputColour, textColour);
    startOriginInput.setValidChars ("0123456789");

    // message boxes and menu

    aboutMessage.setRect (0, 0, 160, 160);	// set "about" size
    aboutMessage.setBackForeColors		// set "about" colours
      (backgroundColour, textColour);
    clearMessage.setBackForeColors		// set "clear" colours
      (backgroundColour, textColour);
    loadMessage.setBackForeColors		// set "load" colours
      (backgroundColour, textColour);
    quitMessage.setBackForeColors		// set "quit" colours
      (backgroundColour, textColour);
    menuBar = mainBar;				// set up main menu
    setMenuBar (menuBar);			// set up menu bar
   }

  // add given boat to scenario if space for it

  private void addBoat (Boat b) {
    if (boatCount < boatMax)			// space for boat?
      boatList[boatCount++] = b;		// add boat to list
  }

  // add rule case to global "caseList"

  public void addCase (Case caseInstance) {
    int i, rule;

    i = 0;
    while (i < RuleBook.NUM_RULES &&
	    (rule = caseInstance.getRule (i)) != 0) {
      if (usedRules[rule])
	caseInstance.suppressRule (i);
      else
	usedRules[rule] = true;
      i++;
    }
    if (caseList == null) {
      caseList = casePointer = caseInstance;
      caseInstance.setNext (null);
    }
    else {
      casePointer.setNext (caseInstance);
      casePointer = caseInstance;
      casePointer.setNext (null);
    }
  }

  // add given mark to scenario if space for it

  private void addMark (Mark m) {
    if (markCount < markMax)			// space for mark?
      markList[markCount++] = m;		// add mark to list
  }

  // add given intention to given boat

  public void addState (Boat b, int intent) {
    switch (intent) {
      case Boat.CLEAR_STATUS:
	removeOverlapped (b);
      case Boat.ABOVE_PROPER_COURSE:
      case Boat.BELOW_PROPER_COURSE:
      case Boat.FINISHING:
      case Boat.GYBING:
      case Boat.STARTING:
      case Boat.TACKING_TO_PORT:
      case Boat.TACKING_TO_STARBOARD:
      case Boat.WAS_AHEAD:
      case Boat.WAS_OVERLAPPED:
	b.setIntention (intent); break;
      case Boat.UNKNOWN:
	break;
      default:
	report ("Unknown status " + intent + "!");
    }
  }

  // make error or alerting sound

  private void beep () {
    Sound.tone (1000, 100);
  }

  // check for boats being too close to each other or a mark

  private void boatsCheck () {
    int i, j;					// boat/mark indexes
    Boat boat1, boat2;				// boats

    for (i = 0; i < boatCount; i++)		// go through boats
      boatList[i].setProximity (false);		// note boat not close
    for (i = 0; i < boatCount; i++) {		// go through boats
      boat1 = boatList[i];			// get first boat
      j = i + 1;				// set second boat index
      while (j < boatCount) {			// go through other boats
	boat2 = boatList[j++];			// get second boat
	if (boat1.isClose (boat2)) {		// boats are close?
	  boat1.setProximity (true);		// note first boat close
	  boat2.setProximity (true);		// note second boat close
	  break;				// leave loop
	}
      }
      j = 0;					// set mark index
      while (j < markCount) {			// go through marks
	if (boat1.isWithin (markList[j++])) {	// boat and mark close close?
	  boat1.setProximity (true);		// note first boat close
	  break;				// leave loop
	}
      }
    }
  }

  // add current scenario to catalogue, returning true/false if
  // added/not added

  private boolean catalogueAdd (String fileName) { // add to catalogue
    Descriptor descriptor;			// descriptor
    int d;					// descriptor index
    return (false);
  }

  // return x/y coordinate, adjusted to map edge if too close to edge

  public int checkEdge (int coord) {
    if (coord < mapEdge)
      coord = mapEdge;
    else if (coord > 160 - mapEdge)
      coord = 160 - mapEdge;
    return (coord);
  }

  // clear boat status list

  private void clearState () {
    editList.remove (apc);
    editList.remove (bpc);
    editList.remove (fin);
    editList.remove (gyb);
    editList.remove (sta);
    editList.remove (tkp);
    editList.remove (tks);
    editList.remove (wad);
    editList.remove (wol);
  }

  // delete boat status

  public void deleteState (Boat b, int intent) {
    switch (intent) {
      case Boat.CLEAR_STATUS:
	b.setIntention (intent);
      case Boat.WAS_OVERLAPPED:
	removeOverlapped (b);
      case Boat.ABOVE_PROPER_COURSE:
      case Boat.BELOW_PROPER_COURSE:
      case Boat.FINISHING:
      case Boat.GYBING:
      case Boat.STARTING:
      case Boat.TACKING_TO_PORT:
      case Boat.TACKING_TO_STARBOARD:
      case Boat.WAS_AHEAD:
	b.cancelIntention (intent); break;
      case Boat.UNKNOWN:
	break;
      default:
	report ("Unknown status " + intent + "!");
    }
  }

  // load scenario with given name

  private void doLoad (String name) {
    if (fileLoad (name)) {
      SailRule.scenarioChange = false;		// note scenario not changed
      filesNameInput.setText ("");
      initialiseMain ();
      drawAfresh (pagePop ());
    }
    else
      report ("No such file");
  }

  // set page to given one and draw it afresh, clearing any report

  private void drawAfresh (int newPage) {
    wipeGUI ();					// remove current components
    page = newPage;				// set new page
    report ("");				// empty report
    drawAgain ();				// draw page again
  }

  // draw page again

  private void drawAgain () {
    if (page == NO_PAGE)			// unknown page?
      page = MAP_PAGE;				// use map as default
    if (page == BOAT_PAGE || page == MAP_PAGE || page == MARK_PAGE ||
	  page == MARK_DIR_PAGE || page == OVERLAPPED_PAGE) {
      this.setBackColor (mapColour);		// colour background for map
      commonReport.setBackColor (mapColour);	// colour background for report
      menuBar =	mainMapBar;			// set up main and map menus
    }
    else {
      this.setBackColor (backgroundColour);	// colour background for other
      commonReport.setBackColor (backgroundColour); // colour backgr. for report
      menuBar =	mainBar;			// set up main menu
    }
    setMenuBar (menuBar);			// set up menu bar
    redraw = true;				// redraw page
    repaint ();					// draw page
  }

  // remove boat from map area

  public void eraseBoat (Boat boat) {
    mapBuffer.fillRect (0, 0, 160, 160);
    boat.drawBoat (mapBuffer);
    mapArea.setDrawOp (Graphics.DRAW_INVERT);
    mapArea.drawImage (mapImage, 0, 0);
    mapArea.setDrawOp (Graphics.DRAW_OVERLAY);
 }

  // remove mark from map area

  public void eraseMark (Mark mark) {
    mapBuffer.fillRect (0, 0, 160, 160);
    mark.drawMark (mapBuffer);
    mapArea.setDrawOp (Graphics.DRAW_INVERT);
    mapArea.drawImage (mapImage, 0, 0);
    mapArea.setDrawOp (Graphics.DRAW_OVERLAY);
 }

  // delete scenario with given name, returning true/false if
  // deleted/not deleted

  public boolean fileDelete (String fileName) {	// delete scenario file
    ObjectCatalog catalogue;			// scenario catalogue
    Descriptor descriptor;			// scenario descriptor
    int d;					// descriptor index
    boolean result = false;			// result
    boolean change = scenarioChange;		// store changed setting

    catalogue =					// open catalogue read-write
      new ObjectCatalog (catalogueName, ObjectCatalog.READ_WRITE);
    catalogue.registerClass (blankDescriptor);	// register descriptor class
    d = fileFind (fileName, catalogue);		// find descriptor index
    if (d != -1) {				// file name found?
      catalogue.deleteObjectAt (d);		// delete current scenario
      filesList.remove (fileName);
      result = true;				// note rename successful
    }
    catalogue.close ();				// close catalogue
    scenarioChange = change;			// restore changed setting
    return (result);				// return delete result
   }

  // find scenario with given file name for given catalogue, returning
  // index/-1 if found/not found

  public int fileFind (				// find descriptor for name
   String fileName, ObjectCatalog catalogue) {
    Descriptor descriptor;			// scenario descriptor
    int d = 0;					// descriptor index

    while ((descriptor =			// read scenario descriptor
	     (Descriptor) catalogue.nextObject ()) != null) {
      if (descriptor.fileName.equals (fileName)) // file name match?
	break;					// stop loop
      d++;					// to next descriptor
    }
    if (descriptor == null)			// descriptor found
      d = -1;					// set invalid index
    return (d);					// return descriptor index
  }

  // make file list from catalogue

  private void fileList () {			// get catalogue file names
    ObjectCatalog catalogue;			// catalogue
    Descriptor descriptor;			// descriptor
    boolean change = scenarioChange;		// store changed setting

    catalogue =					// open catalogue read-only
      new ObjectCatalog (catalogueName, ObjectCatalog.READ_ONLY);
    catalogue.registerClass (blankDescriptor);	// register descriptor class
    while ((descriptor =			// read catalogue descriptor
	    (Descriptor) catalogue.nextObject ()) != null)
      filesList.add (descriptor.fileName);	// store file name
    catalogue.close ();				// close catalogue
    fileSort ();				// make sorted file list
    scenarioChange = change;			// restore changed setting
  }

  // load scenario with given file name,  returning true/false if
  // loaded/not loaded

  public boolean fileLoad (String fileName) {
    ObjectCatalog catalogue;			// scenario catalogue
    Descriptor descriptor;			// scenario descriptor
    int d;					// descriptor index
    boolean result = false;			// result

    initialiseVariables ();			// init scenario variables
    catalogue =					// open catalogue read-only
      new ObjectCatalog (catalogueName, ObjectCatalog.READ_ONLY);
    catalogue.registerClass (blankDescriptor);	// register descriptor class
    d = fileFind (fileName, catalogue);		// find descriptor index
    if (d != -1) {				// descriptor found?

      descriptor =				// load descriptor
	(Descriptor) catalogue.loadObjectAt (d);
      windDir = descriptor.windDir;		// set up wind direction
      boatCount = descriptor.boatCount;		// get boat count
      for (d = 0; d < boatCount; d++)		// for all boats
	boatList[d] = descriptor.boats[d];	// set up boat
      markCount = descriptor.markCount;		// get mark count
      for (d = 0; d < markCount; d++)		// for all marks
	markList[d] = descriptor.marks[d];	// set up mark
      boatsCheck ();				// check boat/mark proximity
      result = true;				// note load successful
    }
    catalogue.close ();				// close catalogue
    return (result);				// return load result
  }

  // rename file from given old name to new name, returning true/false if
  // renamed/not renamed

  public boolean fileRename (			// rename scenario file
   String fileNameOld, String fileNameNew) {
    ObjectCatalog catalogue;			// scenario catalogue
    Descriptor descriptor;			// scenario descriptor
    int d;					// descriptor index
    boolean result = false;			// result
    boolean change = scenarioChange;		// store changed setting

    if (!fileNameOld.equals (fileNameNew) &&	// name differs?
	  filesList.indexOf (fileNameNew) == -1) { // name not in list?
      catalogue =				// open catalogue read-write
	new ObjectCatalog (catalogueName, ObjectCatalog.READ_WRITE);
      catalogue.registerClass (blankDescriptor); // register descriptor class
      d = fileFind (fileNameOld, catalogue);	// find descriptor index
      if (d != -1) {				// file name found?
	descriptor =				// load descriptor
	  (Descriptor) catalogue.loadObjectAt (d);
	catalogue.deleteObjectAt (d);		// delete current scenario
	descriptor.setName (fileNameNew);	// set new file name
	catalogue.addObject ((Storable) descriptor); // add new scenario
	filesList.remove (fileNameOld);		// remove old file name
	filesList.add (fileNameNew);		// add new file name
	fileSort ();				// make sorted file list
	result = true;				// note rename successful
      }
      catalogue.close ();			// close catalogue
    }
    scenarioChange = change;			// restore changed setting
    return (result);				// return rename result
  }

  // save current scenario under given file name, returning true/false if
  // saved/not saved

  public boolean fileSave (String fileName) {	// save scenario
    ObjectCatalog catalogue;			// scenario catalogue
    Descriptor descriptor;			// scenario descriptor
    int d = 0;					// descriptor index
    boolean result = true;			// result

    catalogue =					// open catalogue read-write
      new ObjectCatalog (catalogueName, ObjectCatalog.READ_WRITE);
    catalogue.registerClass (blankDescriptor);	// register descriptor class
    if (catalogue.size () > 0) {		// catalogue non-empty?
      d = fileFind (fileName, catalogue);	// find descriptor index
      if (d != -1)				// file name found?
	catalogue.deleteObjectAt (d);		// delete scenario
      else					// file name not found
	filesList.add (fileName);		// add new file name to list
    }
    else {					// catalogue empty
      catalogue.close ();			// close catalogue
      catalogue =				// open catalogue to create
	new ObjectCatalog (catalogueName, ObjectCatalog.CREATE);
      catalogue.registerClass (blankDescriptor); // register descriptor class
     filesList.add (fileName);			// add new file name to list
    }
    descriptor =				// make scenario descriptor
      new Descriptor (fileName, markList, boatList,
	windDir, boatCount, markCount);
    catalogue.addObject ((Storable) descriptor);// add scenario to catalogue
    catalogue.close ();				// close catalogue
    fileSort ();				// make sorted file list
    return (result);				// return save result
  }

  // make alphabetic file list

  public void fileSort () {			// make file list
    String least, current;			// least name, current name
    int size = filesList.size ();		// get file list size
    int i, j, min;				// indexes

    for (i = 0; i < size; i++) {		// go through file list
      min = i;					// assume index 0 for first
      least = (String) filesList.getItemAt (i);	// get name for least
      for (j = i + 1; j < size; j++) {		// go through other file names
	current = (String) filesList.getItemAt (j); // get other file name
	if (current.compareTo (least) < 0) {	// name earlier than least?
	  min = j; least = current;		// note new least name
	}
      }
      if (min != i) {				// new first found?
	filesList.setItemAt (			// move first to min position
	  min, (String) filesList.getItemAt (i));
	filesList.setItemAt (i, least);		// move least to start
      }
    }
  }

  // get Edit object's text as an angle 0 to 359 degrees, returning
  // -1 if out of range (as only digits, cannot actually be negative)

  private int getAngle (Edit e) {
    int angle = Convert.toInt (e.getText ());
    if (angle < 0 || angle > 359)
      angle = -1;				// angle out of range
    return (angle);
  }

  public final int getBoatCount () {
    return boatCount;
  }

  // return boat list

  public final Boat[] getBoatList () {
    return (boatList);
  }

  // initialise main page

  private void initialiseMain () {
    mainHeadingInput.setText ("");
    mainNameInput.setText ("");
    mainPort.setChecked (false);
    mainPortMark.setChecked (false);
    mainStarboard.setChecked (true);
    mainStarboardMark.setChecked (true);
    voidText ();
  }

  // initialise scenario

  private void initialiseScenario () {
    initialiseMain ();
    initialiseVariables ();
    boatCount = 0;
    markCount = 0;
    pageCount = 0;
    pagePush (MAIN_PAGE);
    pagePush (MAP_PAGE);
    windDir = -1;
    mapArea.setDrawOp (Graphics.DRAW_OVERLAY);
  }

  // initialise key scenario variables

  private void initialiseVariables () {
    scenarioChange = false;			// set scenario not changed
    editPoint = 0;				// set no edit point
    mode = NO_MODE;
    redraw = true;
    repositioning = false;
    state = NORMAL;				// note normal state
    initialiseMain ();
  }

  public int lookupState (String s) {
    Boat b = null;
    int result = -1;
    if (s.equals (apc))
      result = b.ABOVE_PROPER_COURSE;
    else if (s.equals (bpc))
      result = b.BELOW_PROPER_COURSE;
    else if (s.equals (clr))
      result = b.CLEAR_STATUS;
    else if (s.equals (fin))
      result = b.FINISHING;
    else if (s.equals (gyb))
      result = b.GYBING;
    else if (s.equals (sta))
      result = b.STARTING;
    else if (s.equals (tkp))
      result = b.TACKING_TO_PORT;
    else if (s.equals (tks))
      result = b.TACKING_TO_STARBOARD;
    else if (s.equals (wad))
      result = b.WAS_AHEAD;
    else if (s.equals (wol))
      result = b.WAS_OVERLAPPED;
    return (result);
  }

  // handle map keypress

  private void mapKey (int key) {
    switch (key) {
      case 'b': case 'B':			// create boat?
      if (boatCount < boatMax) {		// space for boat?
	  mode = NO_MODE;
	  boatName = Convert.toString (boatCount + 1);
	  if (!nameExists (boatName)) {	// boat name unique?
	    boatHeading = windDir;		// arbitrary heading
	    boatTack = Boat.STARBOARD;	// arbitrary tack
	    drawAfresh (BOAT_PAGE);
	  }
	  else
	    report ("Name " + boatName + " in use");
	}
	else
	  report ("Too many boats");
	break;
      case 'c': case 'C':			// clear scenario?
	if (scenarioChange)			// scenario changed?
	  popupModal (clearMessage);		// popup clear confirmation
	else {					// scenario unchanged
	  initialiseScenario ();		// initialise scenario
	  drawAfresh (START_PAGE);		// redraw screen
	}
	break;
      case 'e': case 'E':			// edit boats?
	mode = NO_MODE;
	if (boatCount > 0) {
	  editPoint = 0;
	  refreshCurrentBoat (editPoint);
	  populateState (boatList[editPoint]);
	  if (pageTop () == EDIT_PAGE)
	    page = pagePop ();
	  else
	    pagePush (MAP_PAGE);
	  drawAfresh (EDIT_PAGE);
	}
	else
	  report ("No boats to edit");
	break;
      case 'f': case 'F':			// files?
	mode = NO_MODE;
	pagePush (MAP_PAGE);
	drawAfresh (FILE_PAGE);
	break;
      case 'h': case 'H':			// change heading?
	if (mode != HEADING_MODE) {
	  mode = HEADING_MODE;
	  prompt ("Drag heading");
	}
	else {
	  mode = NO_MODE;
	  report ("");				// empty report
	  drawAfresh (page);
	}
	break;
      case 'm': case 'M':			// create mark?
	if (markCount < markMax) {		// space for mark?
	  markTack = Boat.STARBOARD;		// arbitrary tack
	  mode = NO_MODE;
	  page = MARK_PAGE;
	  drawAfresh (page);
	}
	else
	  report ("Too many marks");
	break;
      case 'p': case 'P':			// change place?
	if (mode != PLACE_MODE) {
	  mode = PLACE_MODE;
	  prompt ("Drag place");
	}
	else {
	  mode = NO_MODE;
	  report ("");				// empty report
	  drawAfresh (page);
	}
	break;
      case 'r': case 'R':			// display rules?
	mode = NO_MODE;
	page = RULES_PAGE;
	pagePush (MAP_PAGE);
	drawAfresh (page);
	break;
      case 'x': case 'X':			// exit map?
	mode = NO_MODE;
	drawAfresh (pagePop ());
	break;
      default:
	beep ();
    }
  }

  // check if boat name already exists

  private boolean nameExists (String boatName) {
    for (int i = 0; i < boatCount; i++) {
      if ((boatList[i].getName ()).equals (boatName))
	return (true);
    }
    return (false);
  }

  // check if name is empty (i.e. with non-space characters)

  private boolean nonEmpty (String name) {
    int i = 0;
    int length = name.length ();
    char ch;

    while (i < length) {
      ch = name.charAt (i++);
      if (ch != ' ' && ch != '\t' && ch != '\r' && ch != '\n')
	return (true);
    }
    return (false);
  }

  // handle event

  public void onEvent (Event event) {
    int eventType = event.type;

    if (eventType == ControlEvent.WINDOW_CLOSED)
      onEventControl ((ControlEvent) event);
    else if (eventType == KeyEvent.KEY_PRESS)
      onEventKey ((KeyEvent) event);
    else if (eventType == PenEvent.PEN_DOWN)
      onEventDown ((PenEvent) event);
    else if (eventType == PenEvent.PEN_DRAG)
      onEventDrag ((PenEvent) event);
    else if (eventType == PenEvent.PEN_UP)
      onEventUp ((PenEvent) event);
    else if (eventType == ControlEvent.PRESSED)
      onEventPressed (event);
  }

  // handle control event

  public void onEventControl (ControlEvent event) {
    Object target = event.target;		// get event target

    if (target == menuBar) {			// menu bar?
      switch (menuBar.getSelectedMenuItem ()) {	// check menu item
	case 001:				// about
	  popupModal (aboutMessage); break;	// show about message
	case 002:				// quit
	  if (scenarioChange)			// scenario changed?
	    popupModal (quitMessage);		// show quit message
	  else					// scenario not changed
	    exit (0);				// exit program
	  break;
	case 101:				// add boat
	  mapKey ('b'); break;
	case 102:				// add mark
	  mapKey ('m'); break;
	case 103:				// change heading
	  mapKey ('h'); break;
	case 104:				// change place
	  mapKey ('p'); break;
	case 106:				// edit boats
	  mapKey ('e'); break;
	case 107:				// show files
	  mapKey ('f'); break;
	case 108:				// show rules
	  mapKey ('r'); break;
	case 109:				// exit map
	  mapKey ('x'); break;
	case 111:				// clear
	  mapKey ('c'); break;
      }
    }
    else if (target == clearMessage &&		// clear message?
	       clearMessage.getPressedButtonIndex() == 1) { // clear selected?
      initialiseScenario ();			// initialise scenario
      drawAfresh (START_PAGE);			// redraw screen
    }
    else if (target == loadMessage) {		// load message?
      if (loadMessage.getPressedButtonIndex() == 1) { // save selected?
	filesNameInput.setText ("");		// clear current file name
	state = LOADING;			// note loading
	drawAfresh (FILE_PAGE);			// redraw screen
      }
      else					// no save selected
	doLoad (filesNameInput.getText ());	// load file
    }
    else if (target == quitMessage) {		// quit message?
      if (quitMessage.getPressedButtonIndex() == 1) { // save selected?
	filesNameInput.setText ("");		// clear current file name
	state = QUITTING;			// note quitting
	drawAfresh (FILE_PAGE);			// redraw screen
      }
      else					// no save selected
	exit (0);				// exit right away
    }
  }

  // handle pen down event

  public void onEventDown (PenEvent event) {
    int x = event.x;
    int y = event.y;

    if (page == MAP_PAGE) {
      boatSelected = selectedBoat (x, y);
      markSelected = selectedMark (x, y);
    }
  }

  // handle pen drag event

  public void onEventDrag (PenEvent event) {
    float angle;
    int newHeading, tack;
    int xDifference, yDifference, hDifference;
    int x = event.x;
    int y = event.y;

    if (page == MAP_PAGE &&
	  (mode == HEADING_MODE || mode == PLACE_MODE)) {
      if (boatSelected != null) {
	state = DRAGGING;			// note dragging
	eraseBoat (boatSelected);
	if (mode == HEADING_MODE) {
	  xDifference = x - boatSelected.getX ();
	  yDifference = y - boatSelected.getY ();
	  angle = Maths.atan2 ((float) yDifference, (float) xDifference);
	  newHeading = 90 +
	    (int) (angle * 180f / Maths.PI);
	  if (newHeading < 0)
	      newHeading += 360;
	  boatSelected.setHeading (newHeading, windDir);
	  if (!boatSelected.boatOK ())
	    boatSelected.changeTack (windDir);
	}
	else
	  boatSelected.setCoordinates (checkEdge (x), checkEdge (y));
	boatSelected.drawBoat (mapArea);
      }
      else if (markSelected != null) {
	state = DRAGGING;			// note dragging
	eraseMark (markSelected);
	if (mode == HEADING_MODE) {
	  xDifference = x - markSelected.getX ();
	  yDifference = y - markSelected.getY ();
	  angle = Maths.atan2 ((float) yDifference, (float) xDifference);
	  newHeading = 90 +
	    (int) (angle / (2f * Maths.PI) * 360f);
	  if (newHeading < 0)
	    newHeading += 360;
	  hDifference = newHeading - markSelected.getHeading ();
	  if (hDifference < 0)
	    hDifference += 360;
	  if (hDifference >= headingTolerance) { // big enough heading change?
	    if (hDifference < 180)
	      tack = Boat.STARBOARD;
	    else
	      tack = Boat.PORT;
	    markSelected.setHeading (newHeading);
	    markSelected.setTack (tack);
	  }
	}
	else if (mode == PLACE_MODE)
	  markSelected.setCoordinates (checkEdge (x), checkEdge (y));
	markSelected.drawMark (mapArea);
      }
    }
  }

  // handle pen down event

  public void onEventKey (KeyEvent event) {
    int key = event.key;			// get event key

    if (key == IKeys.CALC || key == IKeys.FIND || // calc, find, ...
	key == IKeys.LAUNCH || key == IKeys.HARD1 || // launch, date, ...
	key == IKeys.HARD2 ||  key == IKeys.HARD3 || // phone, todo, ...
	key == IKeys.HARD4)			// memo key?
      popupMenuBar ();				// make menu bar visible
    else if (page == MAP_PAGE)
      mapKey (key);				// handle map key
  }

  // handle control pressed event

  public void onEventPressed (Event event) {
    Boat boat;
    String name;
    int oldHeading, newHeading;
    int status, oldEditPoint;
    Object target = event.target;		// get event target

    // common page widgets
    if (target == commonDone) {
      if (page == EDIT_PAGE) {
	initialiseMain ();
	drawAfresh (pagePop ());
      }
      else if (page == FILE_PAGE &&		// file page?
		 state == QUITTING)		// quitting program?
	exit (0);				// exit program
      else if (page == HELP_PAGE) {
	commonHelp.setText ("Help");
	voidText ();
	drawAfresh (pagePop ());
      }
      else if (page == START_PAGE) {
	windDir = getAngle (startOriginInput);
	if (windDir >= 0) {
	  drawAfresh (pagePop ());
	}
	else {
	  report ("Direction invalid");
	  // drawAgain ();
	}
     }
      else if (page == RULES_PAGE) {
	voidText ();
	drawAfresh (pagePop ());
      }
      else {
	drawAfresh (pagePop ());
      }
    }
    else if (target == commonHelp) {
      if (page == MAIN_PAGE) {
	pagePush (page);
	helpPage = MAIN_PAGE;
      }
      else if (page == HELP_PAGE)
	helpPage = MAP_PAGE;
      else {
	pagePush (page);
	helpPage = page;
      }
      commonText.setFormattedText (help.getText (helpPage));
      commonText.horizontalScroll (0);
      commonText.verticalScroll (0);
      this.setFocus (commonText);
      drawAfresh (HELP_PAGE);
    }

    // edit page widgets
    else if (target == editRedirect) {
      boat = boatList[editPoint];
      oldHeading = boat.getHeading ();
      newHeading = getAngle (editHeadingInput);
      if (newHeading >= 0) {
	boat.setHeading (newHeading, windDir);
	report ("");				// empty report
	if (boat.boatOK ())
	  boatsCheck ();
	else
	  boat.changeTack (windDir);
      }
      else {
	editHeadingInput.setText (Convert.toString (oldHeading));
	report ("Heading invalid");
      }
      this.setFocus (editStatuses);		// statuses get page up/down
    }
    else if (target == editRename) {
      boat = boatList[editPoint];
      boatName = editNameInput.getText ();
      if (!nonEmpty (boatName)) {		// number if no name
	boatName = Convert.toString (editPoint + 1);
	editNameInput.setText (boatName);
      }
      name = boat.getName ();
      if (!boatName.equals (name)) {		// new name differs?
	if (!nameExists (boatName)) {		// boat name unique?
	  renameOverlapped (name, boatName);
	  boat.setName (boatName);
	  report ("");				// empty report
	}
	else
	  report ("Name in use");
      }
      else
	report ("");				// empty report
      this.setFocus (editStatuses);		// statuses get page up/down
    }
    else if (target == editReposition) {
      boat = boatList[editPoint];
      boat.setInvisible ();
      repositioning = true;
      pagePush (EDIT_PAGE);
      drawAfresh (BOAT_PAGE);
    }
    else if (target == editDeleteBoat) {
      boat = boatList[editPoint];
      removeOverlapped (boat);
      if (boatCount == 1) {
	boatList[editPoint] = null;
	boatCount = 0;
	drawAfresh (pagePop ());
      }
      else {
	boatCount--;
	oldEditPoint = editPoint;
	while (editPoint < boatCount)
	  boatList[editPoint] = boatList[++editPoint];
	if (oldEditPoint == boatCount)
	  editPoint = oldEditPoint - 1;
	else
	  editPoint = oldEditPoint;
	boatList[boatCount] = null;
	refreshCurrentBoat (editPoint);
	populateState (boatList[editPoint]);
	if (editPoint == 0)
	  remove (editNext);
	if (editPoint == boatCount - 1)
	  remove (editPrevious);
	boatsCheck ();
	this.setFocus (editStatuses);		// statuses get page up/down
	drawAgain ();
	report ("");				// empty report
      }
    }
    else if (target == editPrevious) {
      if (editPoint > 0)
	editPoint--;
      refreshCurrentBoat (editPoint);
      boat = boatList[editPoint];
      populateState (boatList[editPoint]);
      this.setFocus (editStatuses);		// statuses get page up/down
      report ("");				// empty report
      drawAgain ();
    }
    else if (target == editNext) {
      if (editPoint < boatCount - 1)
	editPoint++;
      refreshCurrentBoat (editPoint);
      boat = boatList[editPoint];
      populateState (boat);
      this.setFocus (editStatuses);		// statuses get page up/down
      report ("");				// empty report
      drawAgain ();
    }
    else if (target == editAddStatus) {
      status = lookupState ((String) editStatuses.getSelectedItem ());
      if (status == Boat.WAS_OVERLAPPED) {
	drawAfresh (OVERLAPPED_PAGE);
      }
      else {
	boat = boatList[editPoint];
	if (status != Boat.UNKNOWN && !boat.getFlag (status)) {
	  addState (boat, status);
	  populateState (boat);
	  editList.repaintNow ();
	}
	report ("");				// empty report
      }
      this.setFocus (editStatuses);		// statuses get page up/down
    }
    else if (target == editClearStatus) {
      boat = boatList[editPoint];
      status = lookupState ((String) editStatuses.getSelectedItem ());
      if (status != Boat.UNKNOWN &&
	   (status == Boat.CLEAR_STATUS || boat.getFlag (status))) {
	deleteState (boat, status);
	populateState (boat);
	editList.repaintNow ();
      }
      report ("");				// empty report
      this.setFocus (editStatuses);		// statuses get page up/down
   }

    // files page widgets
    else if (target == filesList)
      filesNameInput.setText ((String) filesList.getSelectedItem ());
    else if (target == filesDelete || target == filesLoad ||
	       target == filesRename || target == filesSave) {
      name = filesNameInput.getText ();		// get file name
      if (nonEmpty (name)) {			// non-empty file name?
	if (target == filesDelete) {
	  if (fileDelete (name)) {
	    filesNameInput.setText ("");
	    report ("");			// empty report
	    repaint ();
	  }
	  else
	    report ("No such file");
	  this.setFocus (filesList);		// file list gets page up/down
	}
	else if (target == filesLoad) {
	  if (scenarioChange)			// scenario changed?
	    popupModal (loadMessage);		// show load message
	  else					// scenario not changed
	    doLoad (name);			// proceed with load
	}
	else if (target == filesRename) {
	  if (fileRename ((String) filesList.getSelectedItem (), name)) {
	    filesNameInput.setText ("");
	    report ("");			// empty report
	    repaint ();
	  }
	  else
	    report ("Could not rename");
	  this.setFocus (filesList);		// file list gets page up/down
	}
	else if (target == filesSave) {
	  if (fileSave (name)) {		// file saved?
	    if (state == LOADING) {		// loading file?
	      state = NORMAL;			// state now normal
	      scenarioChange = false;		// note scenario not changed
	      drawAfresh (page);		// draw with all buttons
	    }
	    else if (state == QUITTING)		// quitting program?
	      exit (0);				// exit program
	    else {				// not quitting
	      scenarioChange = false;		// note scenario not changed
	      filesNameInput.setText ("");
	      drawAfresh (pagePop ());
	    }
	  }
	  else					// file not saved
	    report ("Could not save");
	}
      }
      else
	report ("Empty file name");
    }

    // main page widgets
    else if (target == mainPort)
      mainStarboard.setChecked (!mainPort.getChecked ());
    else if (target == mainStarboard)
      mainPort.setChecked (!mainStarboard.getChecked ());
    else if (target == mainAddBoat) {
      if (boatCount < boatMax) {		// space for boat?
	boatName = mainNameInput.getText ();
	if (mainPort.getChecked ())
	  boatTack = Boat.PORT;
	else
	  boatTack = Boat.STARBOARD;
	if (!nonEmpty (boatName))		// number if no name
	  boatName = Convert.toString (boatCount + 1);
	if (!nameExists (boatName)) {		// boat name unique?
	  boatHeading = getAngle (mainHeadingInput);
	  if (boatHeading >= 0) {
	    pagePush (MAIN_PAGE);
	    drawAfresh (BOAT_PAGE);
	  }
	  else
	    report ("Heading invalid");
	}
	else
	  report ("Name in use");
      }
      else
	report ("Too many boats");
    }
    else if (target == mainEditBoats) {
      if (boatCount > 0) {
	editPoint = 0;
	refreshCurrentBoat (editPoint);
	populateState (boatList[editPoint]);
	pagePush (MAIN_PAGE);
	drawAfresh (EDIT_PAGE);
      }
      else
	report ("No boats to edit");
    }
    else if (target == mainPortMark)
      mainStarboardMark.setChecked (!mainPortMark.getChecked ());
    else if (target == mainStarboardMark)
      mainPortMark.setChecked (!mainStarboardMark.getChecked ());
    else if (target == mainAddMark) {
      if (markCount < markMax) {		// space for mark?
	if (mainPortMark.getChecked ())
	  markTack = Boat.PORT;
	else
	  markTack = Boat.STARBOARD;
	pagePush (MAIN_PAGE);
	drawAfresh (MARK_PAGE);
      }
      else
	report ("Too many marks");
    }
    else if (target == mainDeleteMark) {
      if (markCount > 0) {
	markList[--markCount] = null;
	boatsCheck ();
	report ("");				// empty report
	drawAgain ();
      }
      else
	report ("No marks to delete");
    }
    else if (target == mainClear)		// clear
      popupModal (clearMessage);		// popup clear confirmation
    else if (target == mainFile) {
      pagePush (MAIN_PAGE);
      drawAfresh (FILE_PAGE);
    }
    else if (target == mainMap) {
      pagePush (MAIN_PAGE);
      drawAfresh (MAP_PAGE);
    }
    else if (target == mainRules) {
      pagePush (MAIN_PAGE);
      drawAfresh (RULES_PAGE);
    }
  }

  // handle pen up event

  public void onEventUp (PenEvent event) {
    Boat boat;
    Mark markTemp;
    int x = checkEdge (event.x);
    int y = checkEdge (event.y);

    // boat page widgets
    if (page == BOAT_PAGE) {
      if (repositioning) {
	boatList[editPoint].setCoordinates (x, y);
	boatList[editPoint].setVisible ();
	repositioning = false;
	page = MAP_PAGE;
	report ("");				// empty report
	boatsCheck ();
      }
      else {
	boat = new Boat (
	  x, y, boatHeading, windDir, boatTack, boatName, null);
	if (boat.boatOK ()) {
	  addBoat (boat);
	  boatsCheck ();
	  report ("");				// empty report
	}
	else
	  report ("Heading impossible");
	page = MAP_PAGE;
      }
      drawAgain ();
    }

    // map page widgets
    else if (page == MAP_PAGE) {
      if (state == DRAGGING) {
	state = NORMAL;
	boatSelected = null;
	markSelected = null;
	boatsCheck ();
	drawAgain ();
      }
   }

    // mark page widgets
    else if (page == MARK_PAGE) {
      markTemp = null;
      markTemp = new Mark (x, y, markTack, windDir);
      addMark (markTemp);
      page = MARK_DIR_PAGE;
      drawAgain ();
    }

    // mark direction page widgets
    else if (page == MARK_DIR_PAGE) {
      markList[markCount - 1].indicateNext (x, y);
      boatsCheck ();
      drawAfresh (MAP_PAGE);
    }

    // overlapped page widgets
    else if (page == OVERLAPPED_PAGE) {
      boat = selectedBoat (x, y);
      if (boat != null && boat != boatList[editPoint]) {
	boatList[editPoint].overlapAdd (boat);
	boat.overlapAdd (boatList[editPoint]);
	addState (boatList[editPoint],
	  lookupState ((String) editStatuses.getSelectedItem ()));
	addState (boat,
	  lookupState ((String) editStatuses.getSelectedItem ()));
	populateState (boatList[editPoint]);
	report ("");				// empty report
      }
      else
	report ("Overlap boat?");
      page = EDIT_PAGE;
      drawAgain ();
    }
  }

  // repaint screen

  public void onPaint (Graphics g) {
    if (redraw) {
      g.setFont (textFont);
      if (page == BOAT_PAGE || page == MAP_PAGE || page == MARK_PAGE ||
	  page == MARK_DIR_PAGE || page == OVERLAPPED_PAGE) {
	for (int i = 0; i < markCount; i++)	// draw boats
	  markList[i].drawMark (mapArea);
	wind.drawWind (windDir, mapArea);	// draw wind
	for (int i = 0; i < boatCount; i++)	// draw marks
	  boatList[i].drawBoat (mapArea);
	if (page == BOAT_PAGE)			// add instructions
	  prompt ("Tap boat position");
	else if (page == MARK_PAGE)
	  prompt ("Tap mark position");
	else if (page == MARK_DIR_PAGE)
	  prompt ("Tap next direction");
	else if (page == OVERLAPPED_PAGE)
	  prompt ("Tap overlap boat");
	add (commonReport);
	this.setFocus (this);			// get keypresses
      }
      else if (page == EDIT_PAGE) {
	add (editEditLabel);
	if (editPoint > 0)
	  add (editPrevious);
	else
	  remove (editPrevious);
	if (editPoint < boatCount - 1)
	  add (editNext);
	else
	  remove (editNext);
	add (editHeadingLabel);
	add (editHeadingInput);
	add (editNameLabel);
	add (editNameInput);
	add (editList);
	add (editRedirect);
	add (editRename);
	add (editReposition);
	add (editDeleteBoat);
	add (editStatuses);
	add (editStatusLabel);
	add (editAddStatus);
	add (editClearStatus);
	add (commonReport);
	add (commonHelp);
	add (commonDone);
	this.setFocus (editStatuses);		// statuses get page up/down
      }
      else if (page == FILE_PAGE) {
	if (filesList.size () == 0) {		// no files loaded?
	  g.setDrawOp (Graphics.DRAW_INVERT);
	  g.drawImage (new Image ("Busy.bmp"), 71, 71);	// draw busy
	  g.setDrawOp (Graphics.DRAW_OVERLAY);
	  fileList ();
	}
	add (filesList);
	add (filesNameInput);
	add (filesNameLabel);
	add (filesSave);
	add (commonReport);
	add (commonHelp);
	add (commonDone);
	if (state != LOADING && state != QUITTING) { // not loading/quitting?
	  add (filesLoad);
	  add (filesRename);
	  add (filesDelete);
	}
	this.setFocus (filesList);		// file list gets page up/down
      }
      else if (page == HELP_PAGE) {
	add (commonScroller);			// add and size scroller
	commonScroller.setRect (0, 0, 159, 145);
	if (helpPage == MAIN_PAGE) {
	  commonHelp.setText ("More");
	  add (commonHelp);
	}
	add (commonDone);
      }
      else if (page == MAIN_PAGE) {
	add (mainBoatLabel);
	mainBoatLabel.setText ("Boats (" + boatCount + "):");
	add (mainNameInput);
	add (mainNameLabel);
	add (mainHeadingLabel);
	add (mainHeadingInput);
	add (mainPort);
	add (mainStarboard);
	add (mainAddBoat);
	add (mainEditBoats);
	add (mainMarkLabel);
	mainMarkLabel.setText ("Marks (" + markCount + "):");
	add (mainPortMark);
	add (mainStarboardMark);
	add (mainAddMark);
	add (mainDeleteMark);
	add (mainFileLabel);
	add (mainClear);
	add (mainFile);
	add (mainMap);
	add (mainRules);
	add (mainWindLabel);
	mainWindLabel.setText ("Wind (" + windDir + ")");
	add (commonReport);
	add (commonHelp);
      }
      else if (page == RULES_PAGE) {
	g.setDrawOp (Graphics.DRAW_INVERT);
	g.drawImage (new Image ("Busy.bmp"), 71, 71); // draw busy
	g.setDrawOp (Graphics.DRAW_OVERLAY);
	rulesOutput (g);			// set up rules
	g.fillRect (0, 0, 160, 160);		// clear busy
	add (commonScroller);			// add and size scroller
	commonScroller.setRect (0, 0, 159, 145);
	add (commonHelp);
	add (commonDone);
      }
      else if (page == START_PAGE) {
	g.setDrawOp (Graphics.DRAW_INVERT);
	g.drawImage (new Image ("Dinghy.bmp"), 39, 3);
	g.setDrawOp (Graphics.DRAW_OVERLAY);
	add (startOriginLabel);
	add (startOriginInput);
	add (commonReport);
	add (commonHelp);
	add (commonDone);
      }
      redraw = false;
    }
  }

  // remove and return page identifier on stack

  private int pagePop () {
    int lastPage;

    if (pageCount > 0)
      lastPage = pageList[--pageCount];
    else
      lastPage = NO_PAGE;			// emergency default
    return (lastPage);
  }

  // add page identifier to stack

  private void pagePush (int page) {
    if (pageCount < pageMax)
      pageList[pageCount++] = page;
  }

  // non-destructively return page identifier on stack

  private int pageTop () {
    int lastPage;

    if (pageCount > 0)
      lastPage = pageList[pageCount - 1];
    else
      lastPage = NO_PAGE;			// emergency default
    return (lastPage);
  }

  public void populateState (Boat b) {
    clearState ();
    boolean tempFlags[] = b.getStates ();
    for (int i = 0; i < tempFlags.length; i++) {
      switch (i) {
	case Boat.ABOVE_PROPER_COURSE:
	  if (tempFlags[i])
	    editList.add (apc);
	  break;
	case Boat.BELOW_PROPER_COURSE:
	  if (tempFlags[i])
	    editList.add (bpc);
	  break;
	case Boat.CLEAR_STATUS:
	  break;
	case Boat.FINISHING:
	  if (tempFlags[i])
	    editList.add (fin);
	  break;
	case Boat.GYBING:
	  if (tempFlags[i])
	    editList.add (gyb);
	  break;
	case Boat.STARTING:
	  if (tempFlags[i])
	    editList.add (sta);
	  break;
	case Boat.TACKING_TO_PORT:
	  if (tempFlags[i])
	    editList.add (tkp);
	  break;
	case Boat.TACKING_TO_STARBOARD:
	  if (tempFlags[i])
	    editList.add (tks);
	  break;
	case Boat.WAS_AHEAD:
	  if (tempFlags[i])
	    editList.add (wad);
	  break;
	case Boat.WAS_OVERLAPPED:
	  if (tempFlags[i])
	    editList.add (wol);
	  break;
	default:
	  report ("Unknown status " + i + "!");
      }
    }
  }

  // prompt message on-screen

  private void prompt (String message) {
    commonReport.setText (message);
  }

  // update details of current boat

  private void refreshCurrentBoat (int editPoint) {
    editNameInput.setText (boatList[editPoint].getName ());
    editHeadingInput.setText (
      Convert.toString (boatList[editPoint].getHeading ()));
  }

  // remove overlap for any boat overlapped on the current one

  private void removeOverlapped (Boat b) {
    int i, j;
    String name1, name2;
    Vector v = b.overlapDelete ();

    scenarioChange = true;			// note scenario change
    for (i = 0; i < boatCount; i++) {
      name1 = boatList[i].getName ();
      for (j = 0; j < v.getCount (); j++) {
	name2 = (String) v.get (j);
	if (name1.equals (name2))
	  boatList[i].overlapRemove (b.getName ());
      }
    }
  }

  // update overlaps with the boat being renamed

  private void renameOverlapped (String oldName, String newName) {
    int i, j;
    String name1, name2;

    for (i = 0; i < boatCount; i++) {
      boatList[i].overlapRename (oldName, newName);
    }
  }

  // report message on-screen

  private void report (String message) {
    if (!commonReport.getText ().equals (message))
      commonReport.setText (message);
    if (!message.equals (""))
      beep ();
  }

  public void rulesCheck () {
    Case caseInstance;
    Mark m;
    Boat a, b;
    int i, j;

    for (i = 0; i < RuleBook.NUM_RULES; i++)
      usedRules[i] = false;
    caseList = casePointer = null;
    if (markCount == 1)
      m = markList[0];
    else
      m = null;
    i = 0;
    while (i < boatCount) {
      a = boatList[i];
      j = i + 1;
      while (j < boatCount) {
	b = boatList[j];
	if (a.nearBoat (b)) {
	  if (m == null) {
	    caseInstance = resolver.openWater (a, b);
	    if (caseInstance != null)
	      addCase (caseInstance);
	  }
	  else {
	    caseInstance = resolver.withMark (m, a, b);
	    if (caseInstance != null)
	      addCase (caseInstance);
	  }
	}
	j++;
      }
      i++;
    }
  }

  public void rulesOutput (Graphics g) {
    String rules = "";
    int instCount = 0;

    rulesCheck ();
    while (caseList != null) {
      if (instCount > 0)
	rules += "\n\n------------------------------------\n\n";
      rules += caseList.stateCase ();
      caseList = caseList.getNext ();
      instCount++;
    }
    commonText.setFormattedText (rules);
    commonText.horizontalScroll (0);
    commonText.verticalScroll (0);
    this.setFocus (commonText);
  }

  // return boat close to given point, or null if none

  private Boat selectedBoat (int x, int y) {
    int i;
    int magX, magY;

    for (i = 0; i < boatCount; i++) {
      magX = Math.abs (boatList[i].getX () - x);
      magY = Math.abs (boatList[i].getY () - y);
      if (magX <= mapTolerance && magY <= mapTolerance)
	return (boatList[i]);
    }
    return (null);
  }

  // return mark close to given point, or null if none

  private Mark selectedMark (int x, int y) {
    int i;
    int magX, magY;

    for (i = 0; i < markCount; i++) {
      magX = Math.abs (markList[i].getX () - x);
      magY = Math.abs (markList[i].getY () - y);
      if (magX <= mapTolerance && magY <= mapTolerance)
	return (markList[i]);
    }
    return (null);
  }

  // clear out stored text and rules

  private void voidText () {
    commonText.setText ("");
    caseList = casePointer = null;
  }

  public void wipeGUI () {
    remove (commonScroller);		// common page widgets
    remove (commonReport);
    remove (commonHelp);
    remove (commonDone);

    if (page == EDIT_PAGE) {		// edit page widgets
      remove (editEditLabel);
      remove (editPrevious);
      remove (editNext);
      remove (editHeadingLabel);
      remove (editHeadingInput);
      remove (editNameLabel);
      remove (editNameInput);
      remove (editList);
      remove (editRedirect);
      remove (editRename);
      remove (editReposition);
      remove (editDeleteBoat);
      remove (editStatuses);
      remove (editStatusLabel);
      remove (editAddStatus);
      remove (editClearStatus);
    }
    else if (page == FILE_PAGE) {	// files page widgets
      remove (filesList);
      remove (filesNameInput);
      remove (filesNameLabel);
      remove (filesLoad);
      remove (filesSave);
      remove (filesRename);
      remove (filesDelete);
    }
    else if (page == MAIN_PAGE) {	// main page widgets
      remove (mainBoatLabel);
      remove (mainNameLabel);
      remove (mainNameInput);
      remove (mainHeadingLabel);
      remove (mainHeadingInput);
      remove (mainPort);
      remove (mainStarboard);
      remove (mainAddBoat);
      remove (mainEditBoats);
      remove (mainMarkLabel);
      remove (mainPortMark);
      remove (mainStarboardMark);
      remove (mainAddMark);
      remove (mainDeleteMark);
      remove (mainFileLabel);
      remove (mainClear);
      remove (mainFile);
      remove (mainMap);
      remove (mainRules);
      remove (mainWindLabel);
    }
    else if (page == START_PAGE) {	// start page widgets
      remove (startOriginLabel);
      remove (startOriginInput);
    }
  }
}
