package jgui.ui;

import waba.ui.*;
import waba.fx.*;
import waba.io.*;
import waba.sys.*;
import waba.util.*;
import extra.ui.*;

/**
 * The <code>Scroller</code> is a container into which the programmer
 * adds a scrollable control.  The control must be an instance of
 * Control and must implement the Scrollable interface.  The scroller
 * manages the display of a vertical and horizontal scrollbar and
 * sizes the scrollable control appropriately.
 *
 * @version $Header: $
 * @author Jim Guistwite
 */

// Ken Turner <kjt@cs.stir.ac.uk>: minor corrections and "setIncrement" method

public class Scroller extends Container implements ScrollbarListener
{
  protected Scrollbar right;
  protected Scrollbar bottom;
  protected Scrollable scrollable;
  private Dimension viewSize;

  /**
   * Construct the scroller using the scrollable provided.  This
   * is the same as calling the empty constructor then calling
   * the setScrollable method.
   * @param s scrollable control (must be a instance of Control and must implement Scrollable).
   */
  public Scroller(Scrollable s) {
    viewSize = new Dimension(0,0);

    right = new Scrollbar(Scrollbar.VERTICAL);
    right.addScrollbarListener(this);
    right.setIncrement(4);
    add(right);

    bottom = new Scrollbar(Scrollbar.HORIZONTAL);
    bottom.setIncrement(4);
    bottom.addScrollbarListener(this);
    add(bottom);

    if (s != null)
      setScrollable(s);
  }

  /**
   * Construct an empty scroller.  You must call setScrollable for
   * the scroller to function correctly.
   */
  public Scroller() {
    this(null);
  }

  /**
   * The invalidate method causes the scrollbars to recompute
   * their state based on the preferred size of the scrollable
   * and the size of the scroller.
   * <P>
   * The programmer should call this if the preferred size of
   * the scrollable changes.  The TextArea class automatically
   * calls this method if it's scroller has been set.
   */
  public void invalidate() {
    sbUpdate();
  }

  /**
   * Set the scrollable associated with this scroller.  The
   * scrollable must be an instance of Control and must implement
   * the Scrollable interface.
   * @param s scrollable
   */
  public void setScrollable(Scrollable s) {
    if (s instanceof Control) {
      scrollable = s;
      add((Control)s);
      scrollable.setViewSize(viewSize.width, viewSize.height);
      //((Control)s).setRect(0, 0, viewSize.width, viewSize.height);
    }
    else {
      // KJT 20/02/02: removed second "-1" parameter
      // Vm.debug("scrollable must be an instance of Control");
    }
    bottom.setIncrement(s.getHorizontalScrollIncrement());
    right.setIncrement(s.getVerticalScrollIncrement());
  }

  //public void paintChildren(Graphics g, int x, int y, int width, int height) {
  //super.paintChildren(g,x,y,width,height);
  //}

  /**
   * Override the setRect method in order to set the sizes of the
   * scrollbars and scrollable.
   * @param x horizontal coordinate.
   * @param y vertical coordinate.
   * @param w width.
   * @param h height.
   */
  public void setRect(int x, int y, int w, int h) {
    super.setRect(x,y,w,h);

    bottom.setRect(0, h-5, w-5, 5);
    right.setRect(w-5, 0, 5, h-5);
    viewSize.height = h - 5;
    viewSize.width = w - 5;

    if (scrollable != null) {
      scrollable.setViewSize(viewSize.width, viewSize.height);
      //((Control)scrollable).setRect(0, 0, viewSize.width, viewSize.height);
      sbUpdate();
    }
  }

  /**
   * Private method to recompute the values of the scrollbar based
   * on the scrollable size and size of this control.
   */
  private void sbUpdate() {
    Dimension sz = scrollable.getPreferredSize();

    if (sz.width > viewSize.width) {
      bottom.setVisible(true);
      bottom.setValues(0,
		       viewSize.width,
		       0,
		       sz.width);
    }
    else {
      bottom.setVisible(false);
    }

    if (sz.height > viewSize.height) {
      right.setVisible(true);
      right.setValues(0,
		      viewSize.height,
		      0,
		      sz.height);
    }
    else {
      right.setVisible(false);
    }
  }

  /**
   * Implement the ScrollbarListener valueChanged method.
   * Pass the new value to the scrollable so that the content
   * can be redrawn.
   * @param sb scrollbar that moved.
   * @param oldVal old value for the scrollbar.
   * @param newVal new value for the scrollbar.
   */
  public void valueChanged(Scrollbar sb, int oldVal, int newVal) {
    if (scrollable != null) {
      if (sb == right)
	scrollable.verticalScroll(newVal);
      else
	scrollable.horizontalScroll(newVal);
    }
  }

  /**
   * Override the onEvent to handle the pilot page up and page
   * down key events.  The page up and page down keys cause the
   * control to scroll up and down.
   * @param e event to be processed.
   */
  public void onEvent(Event e) {
    // KJT 20/08/02: ignore page up/down if no vertical scroll bar visible
    if ((right != null) && (right.visible) && (e.type == KeyEvent.KEY_PRESS)) {
      KeyEvent ke = (KeyEvent)e;
      if (ke.key == IKeys.PAGE_DOWN) {
	right.forward(right.inc);
      }
      else if (ke.key == IKeys.PAGE_UP) {
	right.back(right.inc);
      }
    }
  }

  /**
   * Override findChild so that scrollbars are selected first.
   */
  public Control findChild(int x, int y) {
    Control rv;

    if (right.contains(x, y))
      rv = right;
    else if (bottom.contains(x,y))
      rv = bottom;
    else
      rv = super.findChild(x,y);

    return rv;
  }

  // KJT 06/08/01: Added "setIncrement" method as cannot get at scrollbar

  /**
   * Set the value that should be used to increment or decrement
   * the scrollbar value when the user presses the arrow buttons.
   * @param inc increment value.
   */
  public void setIncrement (int inc) {
    right.setIncrement(inc);
  }
}
